using System;
using System.Collections.Generic;
using System.Linq;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Audio;
using Microsoft.Xna.Framework.Content;
using Microsoft.Xna.Framework.GamerServices;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Input;
using Microsoft.Xna.Framework.Media;
using Microsoft.Xna.Framework.Net;
using Microsoft.Xna.Framework.Storage;

namespace ThatsBetter_NormalMapping
{
    /// <summary>
    /// This is the main type for your game
    /// </summary>
    public class Game1 : Microsoft.Xna.Framework.Game
    {
        GraphicsDeviceManager graphics;
        SpriteBatch spriteBatch;

        KeyboardState lastKbState;

        Model lamp, openbox;

        Effect diffuseColor, normalMappedLight;

        Vector3 cameraPos = new Vector3(1.2f, 1.0f, 1.2f);
        Vector3 lightPosition;

        Texture2D tiles;
        Texture2D tilesNormal;

        Matrix world, view, projection;

        #region MeshDummy

        struct MeshDummy
        {
            public MeshDummy(VertexBuffer vb, VertexDeclaration vd)
            {
                vertexBuffer = vb;
                vertexDeclaration = vd;
            }

            public VertexBuffer vertexBuffer;
            public VertexDeclaration vertexDeclaration;
        };

        Dictionary<ModelMesh, MeshDummy> meshDummies = new Dictionary<ModelMesh, MeshDummy>();
        #endregion

        public Game1()
        {
            graphics = new GraphicsDeviceManager(this);
            Content.RootDirectory = "Content";
        }

        /// <summary>
        /// Allows the game to perform any initialization it needs to before starting to run.
        /// This is where it can query for any required services and load any non-graphic
        /// related content.  Calling base.Initialize will enumerate through any components
        /// and initialize them as well.
        /// </summary>
        protected override void Initialize()
        {
            // TODO: Add your initialization logic here

            base.Initialize();
        }

        /// <summary>
        /// LoadContent will be called once per game and is the place to load
        /// all of your content.
        /// </summary>
        protected override void LoadContent()
        {
            // Create a new SpriteBatch, which can be used to draw textures.
            spriteBatch = new SpriteBatch(GraphicsDevice);

            diffuseColor = Content.Load<Effect>("Diffuse");
            diffuseColor.Parameters["diffuseColor"].SetValue(Color.Gold.ToVector3());
            lamp = Content.Load<Model>("SphereLowPoly");
            ChangeEffect(lamp, diffuseColor);


            normalMappedLight = Content.Load<Effect>("PointLightNormalMap");

            openbox = Content.Load<Model>("OpenBox");

            world = Matrix.Identity;
            view = Matrix.CreateLookAt(cameraPos, -Vector3.One, Vector3.Up);
            float aspectRatio = (float)graphics.GraphicsDevice.Viewport.Width / (float)graphics.GraphicsDevice.Viewport.Height;
            projection = Matrix.CreatePerspectiveFieldOfView(MathHelper.PiOver4, aspectRatio, 0.1f, 100);

            normalMappedLight.Parameters["View"].SetValue(view);
            normalMappedLight.Parameters["Projection"].SetValue(projection);

            normalMappedLight.Parameters["cameraPosition"].SetValue(cameraPos);
            normalMappedLight.Parameters["specularPower"].SetValue(50);
            normalMappedLight.Parameters["specularIntensity"].SetValue(1);

            lightPosition = new Vector3(0, 0, 0);
            normalMappedLight.Parameters["lightPosition"].SetValue(lightPosition);
            SetLightColors(normalMappedLight);

            //--
            AddTangentsToModel(openbox);
            ChangeEffect(openbox, normalMappedLight);

            tiles = Content.Load<Texture2D>("ClassicTileT");
            tilesNormal = Content.Load<Texture2D>("ClassicTileN");
            normalMappedLight.Parameters["ModelTexture"].SetValue(tiles);
            normalMappedLight.Parameters["NormalTexture"].SetValue(tilesNormal);
        }

        private void SetLightColors(Effect effect)
        {
            effect.Parameters["ambientLightColor"].SetValue(new Vector4(0.1f, 0.1f, 0.1f, 1));
            effect.Parameters["diffuseLightColor"].SetValue(new Vector4(0.5f, 0.5f, 0.5f, 1));
            effect.Parameters["specularLightColor"].SetValue(new Vector4(0.5f, 0.5f, 0.5f, 1));
        }

        /// <summary>
        /// UnloadContent will be called once per game and is the place to unload
        /// all content.
        /// </summary>
        protected override void UnloadContent()
        {
            // TODO: Unload any non ContentManager content here
        }

        /// <summary>
        /// Allows the game to run logic such as updating the world,
        /// checking for collisions, gathering input, and playing audio.
        /// </summary>
        /// <param name="gameTime">Provides a snapshot of timing values.</param>
        protected override void Update(GameTime gameTime)
        {
            KeyboardState kbState = Keyboard.GetState();

            // Allows the game to exit
            if (kbState.IsKeyDown(Keys.Escape))
                this.Exit();


            if (kbState.IsKeyDown(Keys.Right))
            {
                lightPosition.X += 0.01f;
                normalMappedLight.Parameters["lightPosition"].SetValue(lightPosition);
            }
            else if (kbState.IsKeyDown(Keys.Left))
            {
                lightPosition.X -= 0.01f;
                normalMappedLight.Parameters["lightPosition"].SetValue(lightPosition);
            }
            if (kbState.IsKeyDown(Keys.PageUp))
            {
                lightPosition.Y += 0.01f;
                normalMappedLight.Parameters["lightPosition"].SetValue(lightPosition);
            }
            else if (kbState.IsKeyDown(Keys.PageDown))
            {
                lightPosition.Y -= 0.01f;
                normalMappedLight.Parameters["lightPosition"].SetValue(lightPosition);
            }
            if (kbState.IsKeyDown(Keys.Down))
            {
                lightPosition.Z += 0.01f;
                normalMappedLight.Parameters["lightPosition"].SetValue(lightPosition);
            }
            else if (kbState.IsKeyDown(Keys.Up))
            {
                lightPosition.Z -= 0.01f;
                normalMappedLight.Parameters["lightPosition"].SetValue(lightPosition);
            }

            lastKbState = kbState;

            base.Update(gameTime);
        }

        /// <summary>
        /// This is called when the game should draw itself.
        /// </summary>
        /// <param name="gameTime">Provides a snapshot of timing values.</param>
        protected override void Draw(GameTime gameTime)
        {
            GraphicsDevice.Clear(Color.CornflowerBlue);

            DrawModel(lamp, Matrix.CreateScale(0.05f) * Matrix.CreateTranslation(lightPosition));
            DrawDummyModel(openbox, world);

            base.Draw(gameTime);
        }

        private void DrawModel(Model model, Matrix world)
        {
            Matrix[] transforms = new Matrix[model.Bones.Count];
            model.CopyAbsoluteBoneTransformsTo(transforms);

            foreach (ModelMesh mesh in model.Meshes)
            {
                foreach (Effect effect in mesh.Effects)
                {
                    effect.Parameters["World"].SetValue(transforms[mesh.ParentBone.Index] * world);
                }
                mesh.Draw();
            }
        }

        public void ChangeEffect(Model model, Effect effect)
        {
            foreach (ModelMesh mesh in model.Meshes)
            {
                foreach (ModelMeshPart part in mesh.MeshParts)
                {
                    part.Effect = effect;
                }
            }
        }

        private void AddTangentsToModel(Model model)
        {
            foreach (ModelMesh mesh in model.Meshes)
            {
                VertexBuffer vb = mesh.VertexBuffer;
                VertexDeclaration vd = mesh.MeshParts[0].VertexDeclaration;
                Tangets.CalculateTangents(mesh.IndexBuffer, ref vb, ref vd, true, true, 0);

                meshDummies.Add(mesh, new MeshDummy(vb, vd));
            }
        }

        public void DrawDummyModel(Model model, Matrix world)
        {
            Matrix[] transforms = new Matrix[model.Bones.Count];
            model.CopyAbsoluteBoneTransformsTo(transforms); 
            
            foreach (ModelMesh mesh in model.Meshes)
            {
                graphics.GraphicsDevice.Indices = mesh.IndexBuffer;
                foreach (ModelMeshPart part in mesh.MeshParts)
                {
                    DrawDummyPart(mesh, part, transforms[mesh.ParentBone.Index] * world);
                }
            }
        }

        public void DrawDummyPart(ModelMesh mesh, ModelMeshPart part, Matrix world)
        {
            part.Effect.Parameters["World"].SetValue(world);

            part.Effect.Begin();

            foreach (EffectPass pass in part.Effect.CurrentTechnique.Passes)
            {
                pass.Begin();

                graphics.GraphicsDevice.VertexDeclaration = meshDummies[mesh].vertexDeclaration;
                graphics.GraphicsDevice.Vertices[0].SetSource(
                    meshDummies[mesh].vertexBuffer,
                    part.StreamOffset,
                    meshDummies[mesh].vertexDeclaration.GetVertexStrideSize(0)
                );

                graphics.GraphicsDevice.DrawIndexedPrimitives(
                    PrimitiveType.TriangleList,
                    part.BaseVertex, 0,
                    part.NumVertices,
                    part.StartIndex,
                    part.PrimitiveCount
                );

                pass.End();
            }

            part.Effect.End();
        }
    }
}
