﻿// Code from:
// http://www.ziggyware.com/readarticle.php?article_id=97


using System;
using System.Collections.Generic;
using System.Linq;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Content;
using Microsoft.Xna.Framework.Graphics;

namespace ThatsBetter_NormalMapping
{
    class Tangets
    {
        public static void CalculateTangents(IndexBuffer IB,
                                     ref VertexBuffer fromVB,
                                     ref VertexDeclaration fromDecl,
                                     bool calcTangent,
                                     bool calcBinormal,
                                     int iStream)
        {
            #region Allocate Tangent and Bitangent
            if ((GetVertexElementByteOffset(fromDecl, VertexElementUsage.Tangent, 0, iStream) == -1 && calcTangent) ||
               (GetVertexElementByteOffset(fromDecl, VertexElementUsage.Binormal, 0, iStream) == -1 && calcBinormal))
            {

                VertexElement[] elems = fromDecl.GetVertexElements();


                VertexElement[] newElems = new VertexElement[elems.Length +
                                                             (calcTangent ? 1 : 0) +
                                                             (calcBinormal ? 1 : 0)];
                elems.CopyTo(newElems, 0);

                if (calcTangent)
                {
                    newElems[elems.Length] =
                        new VertexElement((short)iStream,
                                          (short)fromDecl.GetVertexStrideSize(iStream),
                                          VertexElementFormat.Vector3,
                                          VertexElementMethod.Default,
                                          VertexElementUsage.Tangent,
                                          0);
                }

                if (calcBinormal)
                {
                    newElems[elems.Length + (calcTangent ? 1 : 0)] =
                        new VertexElement((short)iStream,
                                          (short)(fromDecl.GetVertexStrideSize(iStream) +
                                                        (calcTangent ? sizeof(float) * 3 : 0)),
                                          VertexElementFormat.Vector3,
                                          VertexElementMethod.Default,
                                          VertexElementUsage.Binormal,
                                          0);
                }

                VertexDeclaration toDecl = new VertexDeclaration(
                                                fromDecl.GraphicsDevice,
                                                newElems);


                fromVB = ConvertVB(fromVB, fromDecl, iStream, toDecl, iStream);

                fromDecl = toDecl;
            }

            #endregion


            Vector3[] tans = new Vector3[fromVB.SizeInBytes / fromDecl.GetVertexStrideSize(iStream)];
            Vector3[] bitans = new Vector3[fromVB.SizeInBytes / fromDecl.GetVertexStrideSize(iStream)];


            int fromNumVertices = fromVB.SizeInBytes /
                                    fromDecl.GetVertexStrideSize(iStream);

            int newBufferSize = fromNumVertices *
                                    fromDecl.GetVertexStrideSize(iStream);

            byte[] data = new byte[newBufferSize];

            fromVB.GetData<byte>(data);

            int iPosOffset = GetVertexElementByteOffset(fromDecl, VertexElementUsage.Position, 0, iStream);
            int iNormalOffset = GetVertexElementByteOffset(fromDecl, VertexElementUsage.Normal, 0, iStream);
            int iTexOffset = GetVertexElementByteOffset(fromDecl, VertexElementUsage.TextureCoordinate, 0, iStream);
            int iTanOffset = GetVertexElementByteOffset(fromDecl, VertexElementUsage.Tangent, 0, iStream);
            int iBitanOffset = GetVertexElementByteOffset(fromDecl, VertexElementUsage.Binormal, 0, iStream);

            int dwVertSize = fromDecl.GetVertexStrideSize(iStream);

            int numIndices = 0;


            #region 16 bit indices
            if (IB.IndexElementSize == IndexElementSize.SixteenBits)
            {
                numIndices = IB.SizeInBytes / sizeof(short);

                short[] indices = new short[numIndices];
                IB.GetData<short>(indices);

                int count = numIndices / 3;

                for (int i = 0; i < count; i++)
                {
                    float x11 = BitConverter.ToSingle(data,
                            indices[i * 3 + 0] * dwVertSize + iPosOffset + 0 * sizeof(float));
                    float x21 = BitConverter.ToSingle(data,
                            indices[i * 3 + 0] * dwVertSize + iPosOffset + 1 * sizeof(float));
                    float x31 = BitConverter.ToSingle(data,
                            indices[i * 3 + 0] * dwVertSize + iPosOffset + 2 * sizeof(float));

                    Vector3 v1 = new Vector3(x11, x21, x31);

                    x11 = BitConverter.ToSingle(data,
                            indices[i * 3 + 1] * dwVertSize + iPosOffset + 0 * sizeof(float));
                    x21 = BitConverter.ToSingle(data,
                            indices[i * 3 + 1] * dwVertSize + iPosOffset + 1 * sizeof(float));
                    x31 = BitConverter.ToSingle(data,
                            indices[i * 3 + 1] * dwVertSize + iPosOffset + 2 * sizeof(float));

                    Vector3 v2 = new Vector3(x11, x21, x31);

                    x11 = BitConverter.ToSingle(data,
                            indices[i * 3 + 2] * dwVertSize + iPosOffset + 0 * sizeof(float));
                    x21 = BitConverter.ToSingle(data,
                            indices[i * 3 + 2] * dwVertSize + iPosOffset + 1 * sizeof(float));
                    x31 = BitConverter.ToSingle(data,
                            indices[i * 3 + 2] * dwVertSize + iPosOffset + 2 * sizeof(float));

                    Vector3 v3 = new Vector3(x11, x21, x31);

                    x11 = BitConverter.ToSingle(data,
                            indices[i * 3 + 0] * dwVertSize + iTexOffset + 0 * sizeof(float));
                    x21 = BitConverter.ToSingle(data,
                            indices[i * 3 + 0] * dwVertSize + iTexOffset + 1 * sizeof(float));

                    Vector2 w1 = new Vector2(x11, x21);

                    x11 = BitConverter.ToSingle(data,
                            indices[i * 3 + 1] * dwVertSize + iTexOffset + 0 * sizeof(float));
                    x21 = BitConverter.ToSingle(data,
                            indices[i * 3 + 1] * dwVertSize + iTexOffset + 1 * sizeof(float));

                    Vector2 w2 = new Vector2(x11, x21);

                    x11 = BitConverter.ToSingle(data,
                            indices[i * 3 + 2] * dwVertSize + iTexOffset + 0 * sizeof(float));
                    x21 = BitConverter.ToSingle(data,
                            indices[i * 3 + 2] * dwVertSize + iTexOffset + 1 * sizeof(float));

                    Vector2 w3 = new Vector2(x11, x21);


                    float x1 = v2.X - v1.X;
                    float x2 = v3.X - v1.X;
                    float y1 = v2.Y - v1.Y;
                    float y2 = v3.Y - v1.Y;
                    float z1 = v2.Z - v1.Z;
                    float z2 = v3.Z - v1.Z;

                    float s1 = w2.X - w1.X;
                    float s2 = w3.X - w1.X;
                    float t1 = w2.Y - w1.Y;
                    float t2 = w3.Y - w1.Y;

                    float r = 1.0f / (s1 * t2 - s2 * t1);

                    Vector3 sDir = new Vector3((t2 * x1 - t1 * x2) * r,
                                               (t2 * y1 - t1 * y2) * r,
                                               (t2 * z1 - t1 * z2) * r);

                    Vector3 tDir = new Vector3((s1 * x2 - s2 * x1) * r,
                                               (s1 * y2 - s2 * y1) * r,
                                               (s1 * z2 - s2 * z1) * r);

                    tans[indices[i * 3 + 0]] += sDir;
                    tans[indices[i * 3 + 1]] += sDir;
                    tans[indices[i * 3 + 2]] += sDir;

                    bitans[indices[i * 3 + 0]] += tDir;
                    bitans[indices[i * 3 + 1]] += tDir;
                    bitans[indices[i * 3 + 2]] += tDir;



                }



                count = fromVB.SizeInBytes / fromDecl.GetVertexStrideSize(iStream);

                for (int i = 0; i < count; i++)
                {

                    float x11 = BitConverter.ToSingle(data,
                                i * dwVertSize + iNormalOffset + sizeof(float) * 0);
                    float x21 = BitConverter.ToSingle(data,
                                i * dwVertSize + iNormalOffset + sizeof(float) * 1);
                    float x31 = BitConverter.ToSingle(data,
                                i * dwVertSize + iNormalOffset + sizeof(float) * 2);

                    Vector3 n = new Vector3(x11, x21, x31);

                    #region Update Tangent
                    if (iTanOffset != -1)
                    {
                        Vector3 t = tans[i];
                        t = Vector3.Normalize(t - n * (Vector3.Dot(n, t)));

                        byte[] b = BitConverter.GetBytes(t.X);

                        for (int x = 0; x < b.Length; x++)
                        {
                            data[i * dwVertSize + iTanOffset + x + sizeof(float) * 0] =
                                b[x];
                        }

                        b = BitConverter.GetBytes(t.Y);

                        for (int x = 0; x < b.Length; x++)
                        {
                            data[i * dwVertSize + iTanOffset + x + sizeof(float) * 1] =
                                b[x];
                        }

                        b = BitConverter.GetBytes(t.Z);

                        for (int x = 0; x < b.Length; x++)
                        {
                            data[i * dwVertSize + iTanOffset + x + sizeof(float) * 2] =
                                b[x];
                        }


                    }
                    #endregion


                    #region UpdateBitangent
                    if (iBitanOffset != -1)
                    {
                        Vector3 t = bitans[i];

                        t = Vector3.Normalize(t - n * (Vector3.Dot(n, t)));

                        byte[] b = BitConverter.GetBytes(t.X);

                        for (int x = 0; x < b.Length; x++)
                        {
                            data[i * dwVertSize + iBitanOffset + x + sizeof(float) * 0] =
                                b[x];
                        }

                        b = BitConverter.GetBytes(t.Y);

                        for (int x = 0; x < b.Length; x++)
                        {
                            data[i * dwVertSize + iBitanOffset + x + sizeof(float) * 1] =
                                b[x];
                        }

                        b = BitConverter.GetBytes(t.Z);

                        for (int x = 0; x < b.Length; x++)
                        {
                            data[i * dwVertSize + iBitanOffset + x + sizeof(float) * 2] =
                                b[x];
                        }

                    }
                    #endregion

                }

                fromVB.SetData<byte>(data);
            }
            #endregion

            #region 32 bit indices
            if (IB.IndexElementSize == IndexElementSize.ThirtyTwoBits)
            {
                numIndices = IB.SizeInBytes / sizeof(int);

                int[] indices = new int[numIndices];
                IB.GetData<int>(indices);

                int count = numIndices / 3;

                for (int i = 0; i < count; i++)
                {
                    float x11 = BitConverter.ToSingle(data,
                            indices[i * 3 + 0] * dwVertSize + iPosOffset + 0 * sizeof(float));
                    float x21 = BitConverter.ToSingle(data,
                            indices[i * 3 + 0] * dwVertSize + iPosOffset + 1 * sizeof(float));
                    float x31 = BitConverter.ToSingle(data,
                            indices[i * 3 + 0] * dwVertSize + iPosOffset + 2 * sizeof(float));

                    Vector3 v1 = new Vector3(x11, x21, x31);

                    x11 = BitConverter.ToSingle(data,
                            indices[i * 3 + 1] * dwVertSize + iPosOffset + 0 * sizeof(float));
                    x21 = BitConverter.ToSingle(data,
                            indices[i * 3 + 1] * dwVertSize + iPosOffset + 1 * sizeof(float));
                    x31 = BitConverter.ToSingle(data,
                            indices[i * 3 + 1] * dwVertSize + iPosOffset + 2 * sizeof(float));

                    Vector3 v2 = new Vector3(x11, x21, x31);

                    x11 = BitConverter.ToSingle(data,
                            indices[i * 3 + 2] * dwVertSize + iPosOffset + 0 * sizeof(float));
                    x21 = BitConverter.ToSingle(data,
                            indices[i * 3 + 2] * dwVertSize + iPosOffset + 1 * sizeof(float));
                    x31 = BitConverter.ToSingle(data,
                            indices[i * 3 + 2] * dwVertSize + iPosOffset + 2 * sizeof(float));

                    Vector3 v3 = new Vector3(x11, x21, x31);

                    x11 = BitConverter.ToSingle(data,
                            indices[i * 3 + 0] * dwVertSize + iTexOffset + 0 * sizeof(float));
                    x21 = BitConverter.ToSingle(data,
                            indices[i * 3 + 0] * dwVertSize + iTexOffset + 1 * sizeof(float));

                    Vector2 w1 = new Vector2(x11, x21);

                    x11 = BitConverter.ToSingle(data,
                            indices[i * 3 + 1] * dwVertSize + iTexOffset + 0 * sizeof(float));
                    x21 = BitConverter.ToSingle(data,
                            indices[i * 3 + 1] * dwVertSize + iTexOffset + 1 * sizeof(float));

                    Vector2 w2 = new Vector2(x11, x21);

                    x11 = BitConverter.ToSingle(data,
                            indices[i * 3 + 2] * dwVertSize + iTexOffset + 0 * sizeof(float));
                    x21 = BitConverter.ToSingle(data,
                            indices[i * 3 + 2] * dwVertSize + iTexOffset + 1 * sizeof(float));

                    Vector2 w3 = new Vector2(x11, x21);


                    float x1 = v2.X - v1.X;
                    float x2 = v3.X - v1.X;
                    float y1 = v2.Y - v1.Y;
                    float y2 = v3.Y - v1.Y;
                    float z1 = v2.Z - v1.Z;
                    float z2 = v3.Z - v1.Z;

                    float s1 = w2.X - w1.X;
                    float s2 = w3.X - w1.X;
                    float t1 = w2.Y - w1.Y;
                    float t2 = w3.Y - w1.Y;

                    float r = 1.0f / (s1 * t2 - s2 * t1);

                    Vector3 sDir = new Vector3((t2 * x1 - t1 * x2) * r,
                                               (t2 * y1 - t1 * y2) * r,
                                               (t2 * z1 - t1 * z2) * r);

                    Vector3 tDir = new Vector3((s1 * x2 - s2 * x1) * r,
                                               (s1 * y2 - s2 * y1) * r,
                                               (s1 * z2 - s2 * z1) * r);

                    tans[indices[i * 3 + 0]] += sDir;
                    tans[indices[i * 3 + 1]] += sDir;
                    tans[indices[i * 3 + 2]] += sDir;

                    bitans[indices[i * 3 + 0]] += tDir;
                    bitans[indices[i * 3 + 1]] += tDir;
                    bitans[indices[i * 3 + 2]] += tDir;



                }



                count = fromVB.SizeInBytes / fromDecl.GetVertexStrideSize(iStream);

                for (int i = 0; i < count; i++)
                {

                    float x11 = BitConverter.ToSingle(data,
                                i * dwVertSize + iNormalOffset + sizeof(float) * 0);
                    float x21 = BitConverter.ToSingle(data,
                                i * dwVertSize + iNormalOffset + sizeof(float) * 1);
                    float x31 = BitConverter.ToSingle(data,
                                i * dwVertSize + iNormalOffset + sizeof(float) * 2);

                    Vector3 n = new Vector3(x11, x21, x31);

                    #region Update Tangent
                    if (iTanOffset != -1)
                    {
                        Vector3 t = tans[i];
                        t = Vector3.Normalize(t - n * (Vector3.Dot(n, t)));

                        byte[] b = BitConverter.GetBytes(t.X);

                        for (int x = 0; x < b.Length; x++)
                        {
                            data[i * dwVertSize + iTanOffset + x + sizeof(float) * 0] =
                                b[x];
                        }

                        b = BitConverter.GetBytes(t.Y);

                        for (int x = 0; x < b.Length; x++)
                        {
                            data[i * dwVertSize + iTanOffset + x + sizeof(float) * 1] =
                                b[x];
                        }

                        b = BitConverter.GetBytes(t.Z);

                        for (int x = 0; x < b.Length; x++)
                        {
                            data[i * dwVertSize + iTanOffset + x + sizeof(float) * 2] =
                                b[x];
                        }


                    }
                    #endregion


                    #region UpdateBitangent
                    if (iBitanOffset != -1)
                    {
                        Vector3 t = bitans[i];

                        t = Vector3.Normalize(t - n * (Vector3.Dot(n, t)));

                        byte[] b = BitConverter.GetBytes(t.X);

                        for (int x = 0; x < b.Length; x++)
                        {
                            data[i * dwVertSize + iBitanOffset + x + sizeof(float) * 0] =
                                b[x];
                        }

                        b = BitConverter.GetBytes(t.Y);

                        for (int x = 0; x < b.Length; x++)
                        {
                            data[i * dwVertSize + iBitanOffset + x + sizeof(float) * 1] =
                                b[x];
                        }

                        b = BitConverter.GetBytes(t.Z);

                        for (int x = 0; x < b.Length; x++)
                        {
                            data[i * dwVertSize + iBitanOffset + x + sizeof(float) * 2] =
                                b[x];
                        }

                    }
                    #endregion

                }

                fromVB.SetData<byte>(data);
            }
            #endregion
        }

        private static int GetVertexElementByteOffset(VertexDeclaration decl,
                                        VertexElementUsage usage,
                                        byte usageIndex,
                                        int stream)
        {
            VertexElement[] elems = decl.GetVertexElements();
            for (int x = 0; x < elems.Length; x++)
            {
                if (elems[x].Stream == stream &&
                    elems[x].VertexElementUsage == usage &&
                    elems[x].UsageIndex == usageIndex)
                {
                    return elems[x].Offset;
                }
            }
            return -1;
        }

        //-------------------------------
        private static void ConvertVB(ref VertexBuffer vb,
                          VertexDeclaration fromDecl,
                          VertexDeclaration toDecl)
        {
            ConvertVB(vb, fromDecl, 0, toDecl, 0);
        }

        private static void ConvertVB(ref VertexBuffer vb,
                              VertexDeclaration fromDecl,
                              int fromStreamIndex,
                              VertexDeclaration toDecl,
                              int toStreamIndex)
        {
            vb = ConvertVB(vb,
                           fromDecl,
                           fromStreamIndex,
                           toDecl,
                           toStreamIndex);
        }

        private static VertexBuffer ConvertVB(VertexBuffer vb,
                              VertexDeclaration fromDecl,
                              VertexDeclaration toDecl)
        {
            return ConvertVB(vb,
                             fromDecl,
                             0,
                             toDecl,
                             0);
        }

        private static VertexBuffer ConvertVB(VertexBuffer vb,
                              VertexDeclaration fromDecl,
                              int fromStreamIndex,
                              VertexDeclaration toDecl,
                              int toStreamIndex)
        {
            byte[] fromData = new byte[vb.SizeInBytes];
            vb.GetData<byte>(fromData);

            int fromNumVertices = vb.SizeInBytes /
                                    fromDecl.GetVertexStrideSize(0);

            List<int> vertMap = new List<int>();

            //find mappings
            for (int x = 0; x < fromDecl.GetVertexElements().Length; x++)
            {
                VertexElement thisElem = fromDecl.GetVertexElements()[x];

                bool bFound = false;

                int i = 0;
                for (i = 0; i < toDecl.GetVertexElements().Length; i++)
                {
                    VertexElement elem = toDecl.GetVertexElements()[i];

                    if (elem.Stream == toStreamIndex)
                        if (thisElem.VertexElementUsage == elem.VertexElementUsage &&
                            thisElem.UsageIndex == elem.UsageIndex &&
                            thisElem.VertexElementFormat == elem.VertexElementFormat)
                        {
                            bFound = true;
                            break;
                        }
                }
                if (bFound)
                {
                    vertMap.Add(i);
                }
                else
                {
                    vertMap.Add(-1);
                }
            }


            int newBufferSize = fromNumVertices *
                                    toDecl.GetVertexStrideSize(toStreamIndex);



            byte[] toData = new byte[newBufferSize];

            int toDeclVertexStride = toDecl.GetVertexStrideSize(toStreamIndex);
            int fromDeclVertexStride = fromDecl.GetVertexStrideSize(fromStreamIndex);

            for (int x = 0; x < vertMap.Count; x++)
            {
                int i = vertMap[x];

                if (i != -1)
                {
                    VertexElement fromElem = fromDecl.GetVertexElements()[x];
                    VertexElement toElem = toDecl.GetVertexElements()[i];

                    for (int k = 0; k < fromNumVertices; k++)
                    {
                        for (int j = 0; j < GetVertexElementSize(fromDecl, x); j++)
                        {
                            toData[k * toDeclVertexStride + toElem.Offset + j] =
                                fromData[k * fromDeclVertexStride + fromElem.Offset + j];
                        }
                    }
                }
            }

            VertexBuffer newVB = new VertexBuffer(
                vb.GraphicsDevice,
                fromNumVertices * toDecl.GetVertexStrideSize(toStreamIndex),
                BufferUsage.None); // in xna 1.0 use vb.ResourceUsage instead

            newVB.SetData<byte>(toData);

            return newVB;
        }


        //returns the size in bytes of the vertex element given its index
        private static int GetVertexElementSize(VertexDeclaration decl, int elementIndex)
        {
            int elemSize = 0;

            VertexElement[] elems = decl.GetVertexElements();

            if (elems.Length > elementIndex)
            {
                VertexElement e = elems[elementIndex];

                //get the next element if there is one
                int iNextElem = -1;
                if (elems.Length > elementIndex + 1)
                {
                    VertexElement tmp = elems[elementIndex + 1];

                    if (tmp.Stream == e.Stream)
                    {
                        iNextElem = elementIndex + 1;
                    }
                }

                if (iNextElem != -1)
                {
                    elemSize = elems[iNextElem].Offset - e.Offset;
                }
                else
                {
                    elemSize =
                       decl.GetVertexStrideSize(e.Stream) - e.Offset;
                }
            }

            return elemSize;
        }

    }
}
